@@ -9,6 +9,9 @@ import subprocess
 import typing
 import json
 import zipfile
+import shutil
+import logging
+import tqdm
 if sys.version_info[0:2] >= (3, 9):
 	from importlib.resources import files as importlib_resources_files
 else:
@@ -17,9 +20,9 @@ else:
 import shapely.geometry
 
 from fgtools.geo import Rectangle, Coord, get_fg_tile_coords, get_fg_tile_span, get_fg_tile_indices, get_fg_tile_paths, FG_TILE_HEIGHT
-from fgtools import aptdat
+from fgtools import aptdat, get_logger
 from fgtools.utils.files import find_input_files, get_cached_file, get_newest_mtime
-from fgtools.utils import padded_print, read_timestamp, write_timestamp, format_size, download, run_command
+from fgtools.utils import padded_print, read_timestamp, write_timestamp, format_size, download, run_command, quote
 
 GEOFABRIK_REGIONS = {
 	"africa": [
@@ -40,7 +43,7 @@ GEOFABRIK_REGIONS = {
 		'gcc-states', 'india', 'indonesia', 'iran', 'iraq', 'israel-and-palestine', 'japan',
 		'jordan', 'kazakhstan', 'kyrgyzstan', 'laos', 'lebanon', 'malaysia-singapore-brunei', 'maldives',
 		'mongolia', 'myanmar', 'nepal', 'north-korea', 'pakistan', 'philippines',
-		'russian-federation', 'south-korea', 'sri-lanka', 'syria', 'taiwan', 'tajikistan', 'thailand',
+		'south-korea', 'sri-lanka', 'syria', 'taiwan', 'tajikistan', 'thailand',
 		'turkmenistan', 'uzbekistan', 'vietnam', 'yemen'
 	],
 	"australia-oceania": [
@@ -59,11 +62,12 @@ GEOFABRIK_REGIONS = {
 		'germany', 'great-britain', 'greece', 'guernsey-jersey', 'hungary', 'iceland', 'ireland-and-northern-ireland',
 		'isle-of-man', 'italy', 'kosovo', 'latvia', 'liechtenstein', 'lithuania', 'luxembourg', 'macedonia', 'malta',
 		'moldova', 'monaco', 'montenegro', 'netherlands', 'norway', 'poland', 'portugal', 'romania',
-		'russia', 'serbia', 'slovakia', 'slovenia', 'spain', 'sweden', 'switzerland', 'turkey', 'ukraine'
+		'serbia', 'slovakia', 'slovenia', 'spain', 'sweden', 'switzerland', 'turkey', 'ukraine'
 	],
 	"north-america": [
 		'canada', 'greenland', 'mexico', 'us'
 	],
+	"russia": [],
 	"south-america": [
 		'argentina', 'bolivia', 'brazil', 'chile', 'colombia', 'ecuador', 'guyana', 'paraguay', 'peru',
 		'suriname', 'uruguay', 'venezuela'
@@ -118,8 +122,8 @@ OSM_MATERIAL_MAPPINGS = [
 	OsmSelectorPolygon("natural in ('glacier')", "Glacier"),
 	OsmSelectorPolygon("natural in ('mine')", "OpenMining"),
 	OsmSelectorPolygon("natural in ('cliff', 'bare_rock', 'rock', 'scree')", "Rock"),
-	OsmSelectorPolygon("((natural = 'water' and water = '') or water in ('lake', 'oxbow')) and intermittent = 'no' and seasonal = 'no'", "Lake"),
-	OsmSelectorPolygon("((natural = 'water' and water = '') or water in ('lake', 'oxbow')) and (intermittent = 'yes' or seasonal = 'yes')", "IntermittentLake"),
+	OsmSelectorPolygon("((natural = 'water' and water = '') or water in ('lake', 'oxbow')) and (intermittent = 'no' and seasonal is NULL)", "Lake"),
+	OsmSelectorPolygon("((natural = 'water' and water = '') or water in ('lake', 'oxbow')) and (intermittent = 'yes' or seasonal is not NULL)", "IntermittentLake"),
 	OsmSelectorPolygon("natural = 'volcano'", "Lava"),
 	OsmSelectorPolygon("natural = 'scrub'", "Scrub"),
 	OsmSelectorPolygon("natural = 'fell'", "ShrubGrassCover"),
@@ -138,10 +142,10 @@ OSM_MATERIAL_MAPPINGS = [
 	OsmSelectorPolygon("water = 'lagoon'", "Lagoon"),
 	OsmSelectorPolygon("waterway = 'dock' or water = 'harbour'", "Port"),
 	OsmSelectorPolygon("waterway = 'weir'", "BuiltUpCover"),
-	OsmSelectorPolygon("waterway = 'dam'", "Grass"),
+	OsmSelectorPolygon("waterway = 'dam'", "GrassCover"),
 	OsmSelectorPolygon("water = 'canal'", "Canal"),
-	OsmSelectorPolygon("water in ('ditch', 'drain', 'stream_pool') and intermittent = 'no' and seasonal = 'no'", "Stream"),
-	OsmSelectorPolygon("water in ('ditch', 'drain', 'stream_pool') and (intermittent = 'yes' and seasonal = 'yes')", "Stream"),
+	OsmSelectorPolygon("water in ('ditch', 'drain', 'stream_pool') and (intermittent = 'no' and seasona is NULL)", "Stream"),
+	OsmSelectorPolygon("water in ('ditch', 'drain', 'stream_pool') and (intermittent = 'yes' or seasonal is not NULL)", "IntermittentStream"),
 	OsmSelectorPolygon("building != 'yes'", "BuiltUpCover"),
 	OsmSelectorPolygon("highway in ('turning_circle', 'services') or parking = 'site'", "Asphalt"),
 	OsmSelectorLine("highway in ('motorway', 'motorway_link', 'motorway_junction')", "Road", 12),
@@ -153,20 +157,20 @@ OSM_MATERIAL_MAPPINGS = [
 	OsmSelectorLine("highway in ('service', 'track_grade1', 'busway')", "Road", 3),
 	#OsmSelectorLine("highway in ('track', 'track_grade2')", "Gravel", 3),
 	OsmSelectorLine("highway in ('track_grade3', 'track_grade4')", "Sand", 2),
-	OsmSelectorLine("highway = 'track_grade4'", "Grass", 1),
+	OsmSelectorLine("highway = 'track_grade4'", "GrassCover", 1),
 	#OsmSelectorLine("highway = 'track_grade5'", "Dirt", 2),
 	#OsmSelectorLine("highway in ('bridleway', 'path')", "Dirt", 1),
 	OsmSelectorLine("highway in ('footway', 'cycleway')", "Asphalt", 1),
 	OsmSelectorLine("railway in ('rail', 'light_rail')", "Railroad", 1),
 	OsmSelectorLine("railway in ('narrow_gauge', 'tram')", "Railroad", 0.5),
 	OsmSelectorLine("waterway = 'river'", "River", 5),
-	OsmSelectorLine("waterway = 'stream' and intermittent = 'no' and seasonal = 'no'", "Stream", 2),
-	OsmSelectorLine("waterway = 'stream' and (intermittent = 'yes' or seasonal = 'yes')", "IntermittentStream", 2),
+	OsmSelectorLine("waterway = 'stream' and (intermittent = 'no' and seasonal is NULL)", "Stream", 2),
+	OsmSelectorLine("waterway = 'stream' and (intermittent = 'yes' or seasonal is not NULL)", "IntermittentStream", 2),
 	OsmSelectorLine("waterway = 'canal'", "Canal", 10),
-	OsmSelectorLine("waterway = 'ditch' and intermittent = 'no' and seasonal = 'no'", "Stream", 1),
-	OsmSelectorLine("waterway = 'ditch' and (intermittent = 'yes' or seasonal = 'yes')", "IntermittentStream", 1),
-	OsmSelectorLine("waterway = 'drain' and intermittent = 'no' and seasonal = 'no'", "Stream", 0.5),
-	OsmSelectorLine("waterway = 'drain' and (intermittent = 'yes' or seasonal = 'yes')", "IntermittentStream", 0.5),
+	OsmSelectorLine("waterway = 'ditch' and (intermittent = 'no' and seasonal is NULL)", "Stream", 1),
+	OsmSelectorLine("waterway = 'ditch' and (intermittent = 'yes' or seasonal is not NULL)", "IntermittentStream", 1),
+	OsmSelectorLine("waterway = 'drain' and (intermittent = 'no' and seasonal is NULL)", "Stream", 0.5),
+	OsmSelectorLine("waterway = 'drain' and (intermittent = 'yes' or seasonal is not NULL)", "IntermittentStream", 0.5),
 ]
 
 _region_boundary_cache = {}
@@ -196,7 +200,7 @@ def find_region(coord: Coord) -> str:
 			break
 	
 	if not continent:
-		print(f"Warning: found no continent for lon={coord.lon} lat={coord.lat} !")
+		get_logger().warn(f"Warning: found no continent for lon={coord.lon} lat={coord.lat} !")
 		return
 	
 	for region in GEOFABRIK_REGIONS[continent]:
@@ -241,16 +245,17 @@ def find_osm_regions(bboxes: typing.Iterable[Rectangle]):
 	return regions
 
 def download_osm_data(regions: typing.Iterable[str], workspace: str):
-	print("Getting OSM landuse data …")
 	osm_data_folder = os.path.join(workspace, "data", "osm")
 	os.makedirs(osm_data_folder, exist_ok=True)
 	
 	osm_zips = []
 	any_data_files_updated = False
+	pbar = tqdm.tqdm(desc="Downloading OSM landuse data for regions", total=len(regions), unit=" regions")
 	for i, region in enumerate(regions):
 		path = os.path.join(osm_data_folder, region.split("/")[-1] + ".osm.pbf")
 		osm_zips.append(path)
 		old_mtime = os.path.getmtime(path) if os.path.exists(path) else -1
+		pbar.update(1)
 		if not download(
 			GEOFABRIK_DOWNLOAD_URL + region + "-latest.osm.pbf",
 			path,
@@ -283,7 +288,6 @@ def download_osm_data(regions: typing.Iterable[str], workspace: str):
 			any_data_files_updated = True
 
 def download_dem_data(bboxes: typing.Iterable[Rectangle], workspace: str):
-	print("Getting elevation data …")
 	dem_data_folder = os.path.join(workspace, "data", "dem")
 	os.makedirs(dem_data_folder, exist_ok=True)
 	dempkgs = {}
@@ -339,27 +343,29 @@ def process_dem_data(workspace: str, bboxes: typing.Iterable[Rectangle]):
 	timestamp_file = os.path.join(dem_data_folder, "gdalchop")
 	ts = read_timestamp(timestamp_file)
 	for i, bbox in enumerate(bboxes):
-		padded_print(f"Extracting elevation data … {i + 1} of {len(bboxes)} ((n={bbox.top} e={bbox.left} s={bbox.bottom} w={bbox.right})", end="\r")
+		padded_print(f"Chopping elevation data … {i + 1} of {len(bboxes)} ((n={bbox.top} e={bbox.left} s={bbox.bottom} w={bbox.right})", end="\r")
 		tile_paths = list(map(lambda s: s + ".arr.gz", get_fg_tile_paths(bbox)))
 		if get_newest_mtime(hgtfiles) < ts and all(os.path.exists(p) for p in tile_paths) and get_newest_mtime(tile_paths) <= ts:
 			continue
 		
 		any_files_updated = True
-		
-		cmd = f"gdalchop {dem_work_folder} " + " ".join(hgtfiles) + " -- " + " ".join(map(str, get_fg_tile_indices(bbox)))
+		hgtfiles_string = " ".join(map(quote, hgtfiles))
+		cmd = f"gdalchop {quote(dem_work_folder)} {hgtfiles_string} -- {' '.join(map(str, get_fg_tile_indices(bbox)))}"
+		get_logger().debug(f"Running command: {cmd}")
 		p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
 		if p.returncode != 0:
-			log_path = os.path.join(workspace, "log", "dem", os.path.split(hgtfile)[-1])
+			log_path = os.path.join(workspace, "log", "dem", f"N{bbox.top}E{bbox.left}S{bbox.bottom}W{bbox.right}.log")
 			os.makedirs(os.path.join(workspace, "log", "dem"), exist_ok=True)
 			with open(log_path, "wb") as log_file:
 				log_file.write(p.stdout)
-			print(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
+			get_logger().fatal(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
 			sys.exit(1)
-	padded_print(f"Extracting elevation data … {i + 1} of {len(bboxes)} ((n={bbox.top} e={bbox.left} s={bbox.bottom} w={bbox.right})")
+	padded_print(f"Chopping elevation data … {i + 1} of {len(bboxes)} ((n={bbox.top} e={bbox.left} s={bbox.bottom} w={bbox.right})")
 	write_timestamp(timestamp_file)
 	
 	num_arrfiles = len(find_input_files(dem_work_folder, suffix=".arr.gz"))
-	cmd = f"terrafit -m 1000 -x 20000 -e 5 {dem_work_folder}"
+	cmd = f"terrafit -m 1000 -x 20000 -e 5 {quote(dem_work_folder)}"
+	get_logger().debug(f"Running command: {cmd}")
 	p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, text=True, bufsize=1)
 	i = 1
 	cur_path = ""
@@ -378,7 +384,7 @@ def process_dem_data(workspace: str, bboxes: typing.Iterable[Rectangle]):
 			os.makedirs(os.path.join(workspace, "log", "dem"))
 			with open(log_path, "wb") as log_file:
 				log_file.write(p.stdout)
-			print(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
+			get_logger().fatal(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
 			sys.exit(1)
 	padded_print(f"Fitting elevation data … {num_arrfiles} of {num_arrfiles}")
 	
@@ -392,30 +398,22 @@ def process_airports(workspace, aptdat_files):
 	airportarea_folder = os.path.join(work_folder, "AirportArea")
 	os.makedirs(apt_data_folder, exist_ok=True)
 	
-	total_size = sum(map(
-		lambda f: os.stat(f).st_size if os.path.getmtime(f) > os.path.getmtime(os.path.join(apt_data_folder, os.path.basename(f))) else 0,
-		aptdat_files
-	))
-	copied_bytes = 0
-	
 	any_files_updated = False
-	for i, aptdat_file in enumerate(aptdat_files):
-		padded_print(f"Copying apt.dat files … {i + 1} of {len(aptdat_files)} ({os.path.basename(aptdat_file)}) - {format_size(copied_bytes)} of {format_size(total_size)}", end="\r")
-		output_path = os.path.join(apt_data_folder, os.path.basename(aptdat_file))
-		if not os.path.exists(output_path) or os.path.getmtime(aptdat_file) > os.path.getmtime(output_path):
-			any_files_updated = True
-			with open(aptdat_file, "rb") as inf:
-				size = os.stat(aptdat_file).st_size
-				with open(output_path, "wb") as outf:
-					while chunk := inf.read(1000):
-						padded_print(f"Copying apt.dat files … {i} of {len(aptdat_files)} ({os.path.basename(aptdat_file)}) - {format_size(copied_bytes)} of {format_size(total_size)}", end="\r")
-						outf.write(chunk)
-						copied_bytes += len(chunk)
-	padded_print(f"Copying apt.dat files … {len(aptdat_files)} of {len(aptdat_files)} ({os.path.basename(aptdat_file)}) - {format_size(copied_bytes)} of {format_size(total_size)}")
+	
+	genapts_possibilities = ["genapts", "genapts850"]
+	genapts = None
+	for genapts_possibility in genapts_possibilities:
+		if shutil.which(genapts_possibility):
+			genapts = genapts_possibility
+			break
+	
+	if not genapts:
+		get_logger().fatal("No genapts executable found, cannot build airports - exiting !")
+		sys.exit(1)
 	
 	timestamp_file = os.path.join(apt_data_folder, "genapts")
 	ts = read_timestamp(timestamp_file)
-	for i, aptdat_file in enumerate(find_input_files(apt_data_folder, suffix=".dat")):
+	for i, aptdat_file in enumerate(aptdat_files):
 		padded_print(f"Processing apt.dat files … {i + 1} of {len(aptdat_files)} ({os.path.basename(aptdat_file)})", end="\r")
 		if ts > os.path.getmtime(aptdat_file) and os.path.exists(airportarea_folder) and os.path.exists(airportobj_folder) \
 			and get_newest_mtime((airportarea_folder, airportobj_folder)) < ts:
@@ -423,14 +421,15 @@ def process_airports(workspace, aptdat_files):
 		
 		any_files_updated = True
 		
-		cmd = f"genapts --input={aptdat_file} --work={work_folder} --dem-path=dem --max-slope=1"
+		cmd = f"{quote(genapts)} --input={quote(aptdat_file)} --work={quote(work_folder)} --dem-path=dem --max-slope=1"
+		get_logger().debug(f"Running command: {cmd}")
 		p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
 		if p.returncode != 0:
 			log_path = os.path.join(workspace, "log", "apt", os.path.split(aptdat_file)[-1])
 			os.makedirs(os.path.join(workspace, "log", "apt"), exist_ok=True)
 			with open(log_path, "wb") as log_file:
 				log_file.write(p.stdout)
-			print(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
+			get_logger().fatal(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
 			sys.exit(1)
 	padded_print(f"Processing apt.dat files … {len(aptdat_files)} of {len(aptdat_files)} ({os.path.basename(aptdat_file)})")
 	write_timestamp(aptdat_file)
@@ -446,13 +445,14 @@ def ogr_decode(workspace, bbox, cmd, material, work_dir, data_file):
 	env = os.environ.copy()
 	if not "OSM_CONFIG_FILE" in env:
 		env["OSM_CONFIG_FILE"] = importlib_resources_files("fgtools.scenery").joinpath("osmconf.ini")
+	get_logger().debug(f"Running command: {cmd}")
 	p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, env=env)
 	if p.returncode != 0:
 		log_path = os.path.join(workspace, "log", "osm", os.path.split(timestamp_file)[-1])
 		os.makedirs(os.path.join(workspace, "log", "osm"), exist_ok=True)
 		with open(log_path, "wb") as log_file:
 			log_file.write(p.stdout)
-		print(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
+		get_logger().fatal(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
 		sys.exit(1)
 	
 	write_timestamp(timestamp_file)
@@ -462,21 +462,21 @@ def ogr_decode_polygon(workspace, bbox, mapping):
 	osm_data_folder = os.path.join(workspace, "data", "osm")
 	data_file = os.path.join(osm_data_folder, f"N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf")
 	work_dir = os.path.join(workspace, "work", mapping.material)
-	cmd = f"ogr-decode --area-type {mapping.material} --where \"{mapping.selector}\" {work_dir} {data_file} multipolygons"
+	cmd = f"ogr-decode --area-type {quote(mapping.material)} --where {quote(mapping.selector)} {quote(work_dir)} {quote(data_file)} multipolygons"
 	return ogr_decode(workspace, bbox, cmd, mapping.material, work_dir, data_file)
 
 def ogr_decode_line(workspace, bbox, mapping):
 	osm_data_folder = os.path.join(workspace, "data", "osm")
 	data_file = os.path.join(osm_data_folder, f"N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf")
 	work_dir = os.path.join(workspace, "work", mapping.material)
-	cmd = f"ogr-decode --line-width {mapping.line_width} --line-width-column width --area-type {mapping.material} --where \"{mapping.selector}\" {work_dir} {data_file} lines"
+	cmd = f"ogr-decode --line-width {quote(mapping.line_width)} --line-width-column width --area-type {quote(mapping.material)} --where {quote(mapping.selector)} {quote(work_dir)} {quote(data_file)} lines"
 	return ogr_decode(workspace, bbox, cmd, mapping.material, work_dir, data_file)
 
 def ogr_decode_point(workspace, bbox, mapping):
 	osm_data_folder = os.path.join(workspace, "data", "osm")
 	data_file = os.path.join(osm_data_folder, f"N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf")
 	work_dir = os.path.join(workspace, "work", mapping.material)
-	cmd = f"ogr-decode --point-width {mapping.line_width} --area-type {mapping.material} --where \"{mapping.selector}\" {work_dir} {data_file} points"
+	cmd = f"ogr-decode --point-width {quote(mapping.line_width)} --area-type {quote(mapping.material)} --where {quote(mapping.selector)} {quote(work_dir)} {quote(data_file)} points"
 	return ogr_decode(workspace, bbox, cmd, mapping.material, work_dir, data_file)
 
 def process_landuse_data(workspace: str, bboxes: typing.Iterable[Rectangle], regions: typing.Iterable[str]):
@@ -497,8 +497,8 @@ def process_landuse_data(workspace: str, bboxes: typing.Iterable[Rectangle], reg
 			run_command(
 				"osmium extract " +
 					" -b " + ",".join(map(str, (bbox.left, bbox.bottom, bbox.right, bbox.top))) + 
-					" -O -o " + os.path.join(osm_data_folder, f"{os.path.basename(region)}-N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf") +
-					 " " + data_file,
+					" -O -o " + quote(os.path.join(osm_data_folder, f"{os.path.basename(region)}-N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf")) +
+					f" {quote(data_file)}",
 				os.path.join(workspace, "log", "osm", f"osmconvert_{os.path.basename(region)}_N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf")
 			)
 			
@@ -507,25 +507,24 @@ def process_landuse_data(workspace: str, bboxes: typing.Iterable[Rectangle], reg
 		if any_data_files_updated:
 			padded_print(f"Merging extracted OSM data … {i + 1} of {len(bboxes)} (n={bbox.top} e={bbox.left} s={bbox.bottom} w={bbox.right})", end="\r")
 			data_files = [
-				os.path.join(
+				quote(os.path.join(
 					osm_data_folder,
 					f"{os.path.basename(region)}-N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf"
-				)
+				))
 				for region in regions
 			]
 			run_command(
 				"osmium merge " + " ".join(data_files) + " --overwrite -o " +
-					os.path.join(osm_data_folder, f"N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf"),
+					quote(os.path.join(osm_data_folder, f"N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf")),
 				os.path.join(workspace, "log", "osm", f"osmconvert_N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.osm.pbf")
 			)
-			any_files_updated = True
 		
 		landmass_region_file = os.path.join(osm_data_folder, f"landmass-N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.shp")
 		if not os.path.exists(landmass_region_file) or \
 			read_timestamp(landmass_region_file) < max(os.path.getmtime(landmass_file), os.path.getmtime(landmass_region_file)):
 			padded_print(f"Extracting needed OSM data … {i + 1} of {len(bboxes)} (n={bbox.top} e={bbox.left} s={bbox.bottom} w={bbox.right}) from landmass polygons", end="\r")
 			run_command(
-				f"ogr2ogr -clipsrc {bbox.left} {bbox.bottom} {bbox.right} {bbox.top} {landmass_region_file} {landmass_file}",
+				f"ogr2ogr -clipsrc {bbox.left} {bbox.bottom} {bbox.right} {bbox.top} {quote(landmass_region_file)} {quote(landmass_file)}",
 				os.path.join(workspace, "log", "osm", f"ogr2ogr_landmass_N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.shp")
 			)
 			write_timestamp(landmass_region_file)
@@ -546,7 +545,7 @@ def process_landuse_data(workspace: str, bboxes: typing.Iterable[Rectangle], reg
 		
 		data_file = os.path.join(osm_data_folder, f"landmass-N{bbox.top}W{bbox.left}S{bbox.bottom}E{bbox.right}.shp")
 		work_dir = os.path.join(workspace, "work", "Default")
-		cmd = f"ogr-decode --area-type Default {work_dir} {data_file}"
+		cmd = f"ogr-decode --area-type Default {quote(work_dir)} {quote(data_file)}"
 		any_files_updated |= ogr_decode(workspace, bbox, cmd, "Default", work_dir, data_file)
 		
 	padded_print(f"Decoding OSM data … {len(bboxes)} of {len(bboxes)} (n={bbox.top} e={bbox.left} s={bbox.bottom} w={bbox.right})")
@@ -558,13 +557,15 @@ def generate_terrain(workspace: str, bboxes: typing.Iterable[Rectangle], output_
 	num_threads = num_threads or (os.cpu_count() - 1) or 1
 	for i, bbox in enumerate(bboxes):
 			#f"--min-lon={bbox.left} --min-lat={bbox.bottom} --max-lon={bbox.right} --max-lat={bbox.top} " + \
-		cmd = f"tg-construct --threads={num_threads} --output-dir={output_path} --work-dir={work_dir} " + \
+		cmd = f"tg-construct --threads={num_threads} --output-dir={quote(output_path)} --work-dir={quote(work_dir)} " + \
 			"--tile-id=" + " --tile-id=".join(map(str, get_fg_tile_indices(bbox))) + " " + \
-			" ".join(sorted({mapping.material for mapping in OSM_MATERIAL_MAPPINGS}) + ["dem", "AirportArea", "AirportObj", "Default"])
+			" ".join(sorted({quote(mapping.material) for mapping in OSM_MATERIAL_MAPPINGS}) + ["dem", "AirportArea", "AirportObj", "Default"])
+		get_logger().debug(f"Running command: {cmd}")
 		p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, text=True, bufsize=1)
 		j = 0
 		num_tiles = 0
 		for line in p.stdout:
+			get_logger().debug("tg-construct printed: " + str(line.strip()))
 			line = line.strip().split()
 			status_text = ""
 			if len(line) >= 3 and line[2].isnumeric():
@@ -583,7 +584,7 @@ def generate_terrain(workspace: str, bboxes: typing.Iterable[Rectangle], output_
 				os.makedirs(os.path.join(workspace, "log", "terrain"), exist_ok=True)
 				with open(log_path, "w") as log_file:
 					log_file.write(p.stdout.read())
-				print(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
+				get_logger().fatal(f"\nCommand '{cmd}' exited with return code {p.returncode} - see {log_path} for details.")
 				sys.exit(1)
 	
 	padded_print(f"Generating terrain … {len(bboxes)} of {len(bboxes)} (n={bbox.top} e={bbox.left} s={bbox.bottom} w={bbox.right})")
@@ -617,9 +618,22 @@ def main():
 		default=0,
 		type=int
 	)
+	argp.add_argument(
+		"--loglevel",
+		help="Set logging level",
+		default="warning",
+		choices=["debug", "info", "warn", "error", "fatal"]
+	)
 	
 	args = argp.parse_args()
 	
+	args.loglevel = args.loglevel.upper()
+	if hasattr(logging, args.loglevel):
+		loglevel = getattr(logging, args.loglevel)
+	else:
+		loglevel = logging.WARNING
+	get_logger().setLevel(loglevel)
+	
 	aptdat_files = find_input_files(args.input, suffix=".dat")
 	apt_reader = aptdat.ReaderWriterAptDat()
 	apt_reader.read_multiple(aptdat_files)
@@ -643,3 +657,4 @@ def main():
 
 if __name__ == "__main__":
 	main()
+
